/* File yasep/core-js/yasep_asm.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created 2005-06-12 by whygee@f-cpu.org 
2006-08-10 : populated the file (at last)
2006-08-10 : finished the parser and added encoding table/functions (without the flags though)
2006-09-02 : moved to vspsim/JScore/
2006-04-01 : added emit_asm(), set unused register fields,
  tried to enhance the pseudo-instruction parsing
2007-04-04 : update for the Dx pseudo-instructions
2007-04-06 : added check_imm16(i) and 5 new forms for jump/skip
2007-04-08 : cleanup with new JMP/SKIP instruction formats
2007-04-09 : changed the argument for the encoders (form -> flag)
    added the '?' operand, and the X/RX/IX/QX/RRX/IRX/QRX encoders
2007-04-17 : encoded_ignore_fields turned into encoded_fields etc.
2007-04-18 : restricted the range of Imm16 for some instructions
2007-11-05 : rebranding to YASEP
2007-12-25 : MOV 0 -> CLR optimisation 
2008-11-17 : added OPTIONAL|YASEP32_ONLY|YASEP16_ONLY
2009-07-15 : updates
2009-07-19 : small fixes
2009-07-27 : small mask bug with FORM_R,
 objectified the assembler to reduce namespace clutter in the global object.
2009-07-28 : added Y
2009-07-28 : ALIAS_RIR, IgnoreImmSign
2009-07-30 : si4/snd
2009-08-10 : swapped snd & si4 for FORM_RIR, removed SWAPDEST
2009-08-15 : replaced RIR with IRR, added RRR and iRR
2009-08-16 : adding conditional codes
2009-08-23 : removed FORM_X
2011-09-08 : back to work : changed the object structure, create an assembly context,
                put the add_message() as an external function pointer
2011-09-17 : solving the NOP bug by defining NOP=MOV NPC NPC and long NOP forbidden.
               Trying to add IMM20 to MOV
2011-10-17 : bug with CALL and CALL2, IMM size troubles

This file does the assembly (encoding) of instructions
both for interactive debug and coding. It takes
a string of text and spits out an encoded integer
(and other informations) in an object.

 dependency:
   yasep/JScore/txt_int.js,
//////////////  (yasep/JSgui/yasep_messages.js+handling [define an empty function if undesired]),
   yasep/JScore/yasep_forms.js/fields/flags/aliases/conditions
   yasep/JScore/yasep_opcodes.js+children (group_eu.js etc.)
*/


// This creates the object that is passed to and from the assembler
Y.ASM_context=function(add_message, emit_bin) {
  // input
  if (typeof add_message=="function")
    this.add_message = add_message;
  else
    this.add_message = function() {};
  this.YASEP_TYPE=0; // YASEP16/YASEP32 ?

  if (emit_bin)
    this.emit_bin=emit_bin;

  // output :
/* later in the encoder function:
  // preliminary cleanup
  t.encoded_fields=0;       // mask of the used/useful bits
  t.encoded_instruction=0;  // The result of encoding
  t.encoded_size_bytes=0;   // 2 if short form, 4 if long form
  t.asm_operands=[];        // used to pass parameters to external functions
  t.pseudo_flag=0;          // set to 1 if DB/DH/DW is being decoded
  t.X=0; */
}

Y.ASM={

  MAX_FIELD_NR:7, // limit for the parser (except with DB/DH/DW)
    // Should be enough for a complex extended instruction like ADD R1 R2 R3 LSB0 R4

/////////////////////// Common routines ///////////////////////

encode_snd:function(data,t){
  t.encoded_instruction |= data<<Y.OFFSET_SND;
  t.encoded_fields|=Y.FIELD_SND;
  return true
},

encode_si4:function(data,t){
  t.encoded_instruction |= data<<Y.OFFSET_SI4;
  t.encoded_fields|=Y.FIELD_SI4;
  return true
},

encode_dst3:function(data,t){
  t.encoded_instruction |= data<<Y.OFFSET_EXTDST3;
  t.encoded_fields|=Y.FIELD_EXTDST3;
  return true
},

encode_src4:function(data,t){
  t.encoded_instruction |= data<<Y.OFFSET_EXTSRC4;
  return true
},

check_imm16:function(i,flags,t) {
  var a=t.asm_operands[i];
  var c=16;
  // Some instructions don't use the whole field
  if (flags & Y.IMM16_5LSB) {
    if (t.YASEP_TYPE==16) // SHL
      c=4;
    else
      c=5;
  }
  else {
    if (flags & Y.IMM16_8LSB)
      c=8;
  }
  // Add the 4 MSB if IMM20
  if ((flags & Y.IMM20) && (t.YASEP_TYPE!=16)) {
    Y.ASM.encode_snd((a >> 16)&15, t);
    c = 20;
  }

  var d=(1<<c)-1; // build the bitmask
  var b=a&d; // trim the value
  t.encoded_instruction |= 1 | (b<<16);
  t.encoded_fields|=d<<16;
  t.encoded_size_bytes=4;

  if (((a>0)&&(a>d))  // check that no MSB was discarded
  ||  ((a<0)&&(a<-((d+1)>>1))))
    t.add_message('orange', 'Warning : Imm16 too large, keeping only the '+c+' LSB');
  return true
},

check_i:function(i,t) {
  if (t.X==1)
    t.encoded_instruction |= Y.FIELD_EXTIMM4;
  else
    t.encoded_instruction |= 2;
  return Y.ASM.encode_si4(i&15,t)
},

////////////////////////
force_extended_form:function(t){
  t.X=1;
  t.encoded_instruction |= 3;
  t.encoded_size_bytes=4;
  t.encoded_fields |= Y.FIELD_EXTIMM4|Y.FIELD_EXTCOND|Y.FIELD_EXTSRC4;
},


/////////////////////// The form-specific encoders ///////////////////////

encoder_table:[],
  // that table is initialised at the end of the file

/////////////////////// The parser ///////////////////////

// Syntax of an instruction line :
// opcode (operands)* (condition (operands*)) (; comment)

encode_instruction:function(input_text_instruction,t) {
   // arguments : string to encode, ASM context (asm=new ASM_context(add_message))
   // return value : true=OK, false=error

  var i, f=0, g=0, I=false; //I=true when Imm16

  // preliminary cleanup
  t.encoded_fields=0;
  t.encoded_instruction=0;
  t.encoded_size_bytes=0;
  t.asm_operands=[];
  t.pseudo_flag=0;
  t.X=0;

  // lazy preprocessing
  // Character strings are not processed yet !
  var s=input_text_instruction.toUpperCase();
    //  s=s.replace(/[;].*/,'');  // version that breaks minimization
  s=s.replace(new RegExp("[;].*"),'');  // remove any trailing comment
  s=s.replace(/[,]/g,' ');  // remove the commas
  s=s.replace(/\s+/g,' ');  // combine multiple spaces
  s=s.replace(/(^[ ])|([ ]$)/g,''); // strip leading or trailing garbage

  // split the string into "tokens"
  var a=s.split(" ");
  var n=a.length;

  if (a[0]=='') {
    t.add_message('red', 'Error : invalid (empty ?) instruction');
    return false
  }

  /* opcode lookup */
  if (Y.table_opcode[a[0]] == undefined) {

    // Opcode not found. Try a pseudo-instruction :
    t.encoded_size_bytes=0;
    var bytes=Y.table_pseudo_instructions[a[0]];
    if (typeof bytes!='number') {
      t.add_message('red', 'Error : unknown opcode "'+a[0]+'"');
      return false
    }
    else {

      if (a[0]=="YASEP") {
        t.YASEP_TYPE=0;
        t.add_message('green', 'YASEP type reset to default.');
        return true
      }

      if (a[0]=="YASEP16") {
        t.YASEP_TYPE=16;
        t.add_message('green', 'YASEP type set to 16-bit.');
        return true
      }

      if (a[0]=="YASEP32") {
        t.YASEP_TYPE=32;
        t.add_message('green', 'YASEP type set to 32-bit.');
        return true
      }

      else {
        // process DB/DH/DW the pseudo-instructions
        t.pseudo_flag=1;
        t.encoded_instruction=0;

        if (a.length < 2) {
          t.add_message('red', 'Error : missing data for this pseudo-instruction.');
          return false
        }

        // scan the operand list and tries to convert the numbers
        for (var i=1; i<a.length; i++) {
          r=str2int(a[i],'q'); // quiet : don't send an error message
          if (r==null) {
            t.add_message('red', 'Error : invalid number "'+a[i]+'"');
            return false
          }

          // strip the MSB
          var b=0;
          var c=bytes;
          while (c>0) { // stupid way to create a stupid mask
            b= (b<<8) | 0xFF;
            c--;
          }
          b&=r;
          if (b!=r)
            t.add_message('orange', 'Warning : litteral '+a[i]+' is too large. '+(bytes*8)+' LSB kept.');

          if (typeof t.emit_bin == 'function')
            t.emit_bin(b, bytes);

          // This works for data <= 32 bits total :
          if (t.encoded_size_bytes<4)
            t.encoded_instruction+= (b<<(t.encoded_size_bytes*8));
          t.encoded_size_bytes+=bytes;
        }
        t.add_message('blue',' OK, '+t.encoded_size_bytes+' literal bytes.');
      }
    }
  }
  // normal instruction :
  else {
    var to=Y.table_opcode[a[0]];
    t.encoded_instruction=to.value;
    t.encoded_fields=Y.FIELD_OPCODE|Y.FIELD_EXTFORM|Y.FIELD_IMMREG;
    t.encoded_size_bytes=2; // Default. size=4 is set in check_imm16()
    f=to.forms;
    g=to.flags;
    var p='';

// Parse the operands and build the form :

    // decode the other fields (speculatively)
    for (i=1; i<n; i++) {
      // first try with a register
      var r=Y.names_registers[a[i]];
      if (typeof r == 'number') {
        p=p+'R';
        t.asm_operands[i]=r;
      }
      else {
        // is it a number ?
        r=str2int(a[i],'q'); // quiet : don't send an error message
        if (typeof r == 'number') {
          var c='I';
          if ((r >= -8) && ((r <= 7)
          || ((r >=  0) && ( r <= 15) && (g&Y.IgnoreImmSign))) && (f & Y.FORMS_i))
            c='i';
          else
            I=true;
          p=p+c;
         t.asm_operands[i]=r;
        }
        else {
          var k=Y.conditionTable[a[i]];
          if (typeof k == 'object') {
            if (I==true) {
              t.add_message('red',"Param #"+i+", Error : conditional instructions can't hold long immediate values");
              return false
            }
            if (g & Y.NO_CONDITION)
              t.add_message('orange',"Warning : this opcode is not interruptible and does not accept conditions. Continuing anyway.");

            Y.ASM.force_extended_form(t);
            t.encoded_instruction|= k.val<<Y.OFFSET_EXTCOND;
            var l=k.reg;
            if (typeof l=='number')
              Y.ASM.encode_src4(l,t);
            else {
              if (typeof l=='string')
                Y.ASM.encode_src4(Y.names_registers[l],t);
              else {
                i++;
                r=Y.names_registers[a[i]];
                if (typeof r == 'number')
                  Y.ASM.encode_src4(r,t);
                else {
                  t.add_message('red', 'Error : a register operand is expected after this condition');
                  return false
                }
              }
            }
            // complain if there is stuff after this :
            if (i+1 != n) {
              t.add_message('red', 'Error : too many operands after the condition');
              return false
            }
          }
          else {
            t.add_message('red', 'Error : invalid operand: "'+a[i]+'"');
            return false
          }
        }
      }
    } // end for() loop

    if (p=='') // no operand ?
      p='FORM_ALONE';
    else
      p='FORM_'+p;

    if (typeof Y[p] != 'number') {
      // try with a long immediate
      p=p.replace(/i/g,'I'); // Note : it may not work if more than 1 immediate field is implemented in YASEP
      if (typeof Y[p] != 'number') {
        t.add_message('red', 'Error : operand combination  "'+p+'" does not exist');
        return false
      }
    }

    if ((t.YASEP_TYPE==32) && (g & Y.YASEP16_ONLY)) {
      t.add_message('orange', 'Warning : opcode "'+a[0]+'" is valid only in 16-bit mode');
    } else {
      if ((t.YASEP_TYPE==16) && (g & Y.YASEP32_ONLY)) {
        t.add_message('orange', 'Warning : opcode "'+a[0]+'" is valid only in 32-bit mode');
      }
    }

    // Now we have one entry in the bitmask, check if it is present in the opcode
    var mask=f & Y[p]; // form lookup
    if (mask == 0) {
      t.add_message('orange', 'Error : operand combination  "'+p+'" is not valid for opcode "'+a[0]+'"');
      return false
    }

    // call the form-specific encoder
    if (typeof Y.ASM.encoder_table[mask]!='function') {
      t.add_message('red', 'Internal error : no encoder for form '+Y.TableForm[mask]);
      return false
    }
    if (!Y.ASM.encoder_table[mask](g,t)) // relegate the rest of the work to specific routines
      return false;

    // Verify that something has been really processed
    if (~(t.encoded_fields) & t.encoded_instruction)
       t.add_message('orange','Warning : (internal bug) "encoded_fields" is missing bits'
        + ' encoded_fields:'+int2hex(t.encoded_fields)
        +' instruction:'+int2hex(t.encoded_instruction));
  
    // the end :
    t.add_message('blue',' OK: opcode='+Y.decorate_opcode(t.encoded_instruction & Y.FIELD_OPCODE)
         +' form='+p+' '+t.encoded_size_bytes+' bytes mask='
         +int2hex(t.encoded_fields,(t.encoded_size_bytes*2)-1,(t.encoded_size_bytes*2)-1));
    if (typeof t.emit_bin == 'function')
      t.emit_bin(t.encoded_instruction, t.encoded_size_bytes);
//    else
//      if (typeof emit_bin == 'function')
//        emit_bin(t.encoded_instruction, t.encoded_size_bytes);
  } // instruction
  return true
}

}; /* Y.ASM() */


//////////////////////////////////////////////////////////////////////////
/////////////////////// The form-specific encoders ///////////////////////

Y.ASM.encoder_table[Y.FORM_ALONE]= // NOP, HALT, INV...
function() {
  /* nothing */
  return true
};

Y.ASM.encoder_table[Y.FORM_R]=
function(flag,t) {
// only used by :
// NEG = 0 - snd
// NOT = 0 | ~snd
//  if (flag & Y.ALIAS_RR) ?
  t.check_i(0,t);
  if (t.X==1)
    Y.ASM.encode_dst3(t.asm_operands[1],t);
  else
    Y.ASM.encoded_instruction |= 2; // force FORM_iR
  Y.ASM.encode_snd(t.asm_operands[1],t);
  return true
};

Y.ASM.encoder_table[Y.FORM_RR]=  // used by ROP2, ASU, SHL...
function(flag,t) {
  // NEG = 0 - src1
  // NOT = 0 | ~src1
  if (flag & Y.ALIAS_RR) {
    Y.ASM.force_extended_form(t);
    Y.ASM.encode_snd(t.asm_operands[1],t);
    Y.ASM.check_i(0,t);
    Y.ASM.encode_dst3(t.asm_operands[2],t);
  }
  else {
    Y.ASM.encode_si4(t.asm_operands[1],t);
    if (t.X==1)
      Y.ASM.encode_dst3(t.asm_operands[2],t);
//    else ?
      Y.ASM.encode_snd(t.asm_operands[2],t);
  }
  return true
};

Y.ASM.encoder_table[Y.FORM_RRR]=  // another great classic...
function(flag,t) {
  Y.ASM.force_extended_form(t);
  Y.ASM.encode_si4 (t.asm_operands[1],t);
  Y.ASM.encode_snd (t.asm_operands[2],t);
  Y.ASM.encode_dst3(t.asm_operands[3],t);
  return true
};

/*************** With i ***************/

Y.ASM.encoder_table[Y.FORM_i]=
function (flag,t) {
  // used only by CRIT which is not abortable and has no destination
  // so no need to deal with the extended form.
  return Y.ASM.check_i(t.asm_operands[1],t)
};

Y.ASM.encoder_table[Y.FORM_iR]=
function (flag,t,X) {
  if (t.X==1)
    Y.ASM.encode_dst3(t.asm_operands[2],t);
  Y.ASM.encode_snd(t.asm_operands[2],t);
  return Y.ASM.check_i(t.asm_operands[1],t)
};

Y.ASM.encoder_table[Y.FORM_Ri]=
function (flag,t) {
  // Used only by PUT so no need to deal with dst3
  return (Y.ASM.check_i(t.asm_operands[2],t) && Y.ASM.encode_snd(t.asm_operands[1],t))
};

Y.ASM.encoder_table[Y.FORM_iRR]=  // extended form
function(flag,t) {
  Y.ASM.force_extended_form(t);
  Y.ASM.encode_snd (t.asm_operands[2],t);
  Y.ASM.encode_dst3(t.asm_operands[3],t);
  return Y.ASM.check_i(t.asm_operands[1],t)
};

/*************** With I ***************/

Y.ASM.encoder_table[Y.FORM_RI]=  // PUT
function (flag,t) {
  return (Y.ASM.check_imm16(2,flag,t) && Y.ASM.encode_snd(t.asm_operands[1],t))
};

Y.ASM.encoder_table[Y.FORM_IR]=
function (flag,t) {
  if(!Y.ASM.check_imm16(1,flag,t))
    return false;
  if (flag & Y.SWAP_IR) // GET, MOV
    return Y.ASM.encode_si4(t.asm_operands[2],t);
  if (flag & Y.ALIAS_IRR) // ADD...
         Y.ASM.encode_si4(t.asm_operands[2],t);
  return Y.ASM.encode_snd(t.asm_operands[2],t)
};

Y.ASM.encoder_table[Y.FORM_IRR]=  // used by SHL, ROP2, ASU
function (flag,t) {
  if(!Y.ASM.check_imm16(1,flag,t))
    return false;

         Y.ASM.encode_snd(t.asm_operands[2],t);
  return Y.ASM.encode_si4(t.asm_operands[3],t)
}

