/*
yasep/gui-js/winlist.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created sam. août 13 22:06:58 CEST 2011 by whygee@f-cpu.org
version mar. sept. 13 01:42:52 CEST 2011
version mer. nov.  9 13:40:47 CET 2011 : add counter to the delay, renamed from ygwm-winlist.js to winlist.js
           and add color-grouping

creates and maintains a foldable vertical list of open windows.
this is also the official list of all the open windows so 
that's where they are accessed and it affects the cookies too.

DOM tree :
body->winlist->header->minimize
             |       L>"Windows"
             L>winlist_bg->winlist_fg->Winlist-item
                                     L>Winlist-item
                                     L>Winlist-item
                                     ...
CSS:
winlist:
   z-index: 1000
   defined width
   has borders, background
   position: absolute at the bottom left of the page, "fixed" (moves when scrolling)
header:
   like ygwm (reuse styles)
winlist_bg:
   determines the size (height+width) of the list,
   shown/hidden to fold/unfold the list
winlist_fg:
   contains the items,
   height left to the browser and read to set the height of winlist_bg

20110829 : solved the winlist height/resize issue by adding winlist_fg,
    added colors to the items, started support of winlist.list
*/

init_obj.push(winlist={
  margin_top:80,

//  list:{}, /* The list of the windows, initialised by create */

  change_window:function(win,item){
    if (item.className=="winlist_item") { // not selected=> select
      ygwm.focus_window(null,win);
      if (win.style.display=="none")
        win.show_all();
    }
    else {
      if (win.style.display=="none")
        win.show_all();
      else
        return win.hide_all(); // selected => hide
    }

    // in the first 2 cases, scroll to the window
    ygwm.center_window(win);
  },

  // delaying solved both a CPU usage concern and a display init issue.
  adjust_height_obj:{
    delay:200, // ms
    max:6,
    counter:6,
    timer:null,
    func:function(){
      var y=winlist.winlist_fg.scrollHeight,
          z=getHeight();

      if ((y+winlist.margin_top) > z) { // clip
        y = 0;
        z -= winlist.margin_top;
        if (z >= 0)
          y = z;
      }
      winlist.winlist_bg.style.height = y+'px';
    }
  },
  adjust_height:function(){
    delay(winlist.adjust_height_obj);
  },

  MinMaxClick: function(){
    if (winlist.minmax_button.className=="ygwm_corner_minimize"){
      winlist.winlist_bg.style.display="none";
      winlist.minmax_button.className="ygwm_corner_maximize";
    }
    else {
      winlist.winlist_bg.style.display="";
      winlist.minmax_button.className="ygwm_corner_minimize";
    }
    winlist.adjust_height(); // just in case.
  },

  removeItem:function(win){
    winlist.list.splice(win.winlist_index,1); // remove from list
    winlist.winlist_fg.removeChild(win.winlist_item_id);
    winlist.adjust_height();
  },

  rename:function(item,val){
    item.innerHTML=val;
    winlist.adjust_height();
  },

  addWindow:function(win,name,style){
    var h=document.createElement("DIV");
    win.really_hide=true;
    win.winlist_item_id=h; // so ygwm.focus_window() can change h.className
//    h.className="winlist_item"; // will be changed by the focus_window() in ygwm
    h.innerHTML=name;
    h.win=win;
    h.onclick=function(){winlist.change_window(win,h);};
//    if (YGCSS.ColorStyles[style]) // moved to focus_window()
//      h.style.backgroundColor=YGCSS.ColorStyles[style][5];
    winlist.winlist_fg.appendChild(h);

    win.winlist_index=winlist.list.length; // register then add to list
    winlist.list[winlist.list.length]=win;

    winlist.adjust_height();
  },

  minimizeAll: function(){
    for (var i in winlist.list)
      winlist.list[i].hide_all();
  },

  restoreAll: function(){
    for (var i in winlist.list)
      winlist.list[i].show_all();
  },

  init_obj:function(){
    if(winlist.winlist)
      return alert("winlist already started");
    if(!ygwm.new_window)
      return alert("ygwm not working");

    winlist.list=[];

    // main DIV
    var w=document.createElement("DIV");
    w.className="ygwm_div";  // winlist reuses the ygwm CSS but we have to adapt it "a bit"
    w.id="winlist";  // add more style from in ygwm.css
    w.innerHTML=
       '<div class="ygwm_header_selected">'
      +'<div class="ygwm_corner_minimize"></div>'
      +'<div id="winlist_name">Windows list</div>'
      +'</div>'
      +'<div id="winlist_bg"><div id="winlist_fg"></div></div>';
    document.body.appendChild(w);
    winlist.winlist=w;
    
    winlist.header=getByClass("ygwm_header_selected","DIV",w); // header DIV
    winlist.header.onclick=winlist.MinMaxClick;
    winlist.minmax_button=getByClass("ygwm_corner_minimize","DIV",winlist.header);    // minimize/maximize button
    winlist.winlist_bg=document.getElementById("winlist_bg");    // background of the list
    winlist.winlist_fg=document.getElementById("winlist_fg");    // foreground of the list

    window.onresize=winlist.adjust_height; // the vertical changes should be filtered
  }
});
